Re:  Recursive Programming
Dt:  June 15, 1993


Recursive procedure calls is a feature most low level programming
languages provide.  Like a dog chasing its tail, a recursive
procedure executes itself over and over.  One problem a language
like C could experience is 'stack overflow' (error 1308 in
FoxPro) or the program can run out of allocated memory(error 1282
in FoxPro).

FoxPro also has a limit of 32 'DO' program nesting levels (error
103) that can easily be exceeded when performing recursion.  We
can simulate this error with the following program:

     clear
     cnt = 0
     do recurtmz    &&call the recursive procedure

     procedure recurtmz
     cnt = cnt + 1
     ? cnt
     do recurtmz    &&call it again...

The program will abort after the number 30 has been displayed to
the screen.  Now that we are aware that there are some
limitations to recursive programming, let's write a beneficial
program that exploits its power.

Our program is similar to FileFind in FoxPro's FILER.  It will
search a directory and all subordinate directories for files that
meet a specific search pattern.  All files that meet the search
pattern will be listed to the screen and also saved to a database
table.  After the search is complete, all files that met the
search pattern will be presented to the user in a browse window. 
If the user presses the ENTER key within the browse window, the 
file will be opened for viewing. 

Before we begin, we need to create a database called FILES.DBF
that contains the following fields:

     Fieldname      Type      Length    Decimals
     =========      ====      ======    ========
     FFILE           C          12
     FSIZE           N          10          0
     FDATE           D
     FTIME           C           8
     FDIR            C          65

FDIR is set to 65 characters since the maximum length of an MS-
DOS path is 63 characters.  We'll add 2 more characters for the
drive letter and colon.

Our main procedure is contained in lines 1 through 40.  We'll
first call the GETDIR() function to select which directory is to
be searched.  If a directory is selected, we'll open the FILES
table and zap it so that all new files found won't be mixed in
with previous searched files.   


The operator will be asked to enter a file name search pattern
(*.PRG). We also will provide a get field for storing a
'contains' specification.  This will allow us to specify a word
that must be contained in the file name.  If our search pattern
is *.PRG and the word TEST was entered into the contains field,
files like ATEST.PRG, TEST.PRG, ATEST1A.PRG, etc would be
selected.

The recursive procedure, FINDFILE, is then called and the
directories are searched.  If any files are found, they are
presented to the operator in a browse window and may be viewed. 

  1. set talk off
  2. release all
  3. clear
  4. 
  5. cDir_select = getdir()        && directory name to search 
  6. 
  7. if not empty(cDir_select)     && run if a directory was selected
  8. 
  9.      set safety off
 10.           use files exclusive
 11.           zap
 12.      
 13.           *these are our global variables
 14.           cDirspec  = "*."
 15.           cFilespec = "*.*" + space(9)  && length of maximum file name
 16.           cContain  = space(8)          && does filename contain this?
 17. 
 18.           *get our search criteria
 19.           @ 1,1 say "File spec:     " get cFilespec function "!"
 20.           @ 2,1 say "File Contains: " get cContain  function "!"
 21.           read
 22. 
 23.           *find the files 
 24.           if not empty(cFilespec)
 25.                do filefind with cDir_select
 26.           endif     
 27.      
 28.           *view the results
 29.           if reccount() # 0
 30.                go top
 31.                on key label ENTER do viewfile     &&view the file
 32.                browse nomodify
 33.                on key label ENTER            &&release ON KEY
 34.           endif
 35.           
 36.           close data
 37.      
 38. endif     
 39. 
 40. return

The directory name to be searched is passed to the FileFind
procedure.  We'll use private variables so that the values do not
conflict with variables or the same name from prior calls.

 41. procedure filefind
 42. parameter cCurdir             && directory to start from
 43. 
 44. private aDirname              && array to hold directory name
 45. private aFilename             && array to hole file names
 46. private nRowdir               && number of directories in aDirname
 47. private nRowfile              && number of files in aFilename
 48. private nFor                  && for-loop counter
 49. private lDir                  && output the directory name?
 50. private lFile            && output the file name?
 

We'll store the directory names in one array and the file names
in a separate array.  If no directories meet the directory
pattern ("*.") or no files meet the filename search pattern, no
array will be created and the number of rows in to that array
name will be 0.  

 51. =adir(aDirname,cCurdir +  cDirspec,"D")
 52. =adir(aFilename,cCurdir + cFilespec)
 53. 
 54. nRowdir  = iif(type("aDirname") # "U",alen(aDirname,1),0)
 55. nRowfile = iif(type("aFilename") # "U",alen(aFilename,1),0)


We will first list all files in the directory we are searching in
that meet the search pattern.  If a 'contains' clause was
entered, we'll verify the file name contains that clause.  

The first time a file in a directory meets all required search
patters, the directory name will be echoed to the screen.  All
file names meeting that pattern will then be echoed to the screen
and appended into our table.  When all file names have been
searched in the array, the file name array is released to
conserve memory.

 56. lDir = .T.
 57. 
 58. for nFor = 1 to nRowfile
 59.      if substr(aFilename(nFor,5),5,1) # "D"
 60.           lFile = .T.
 61.           if not empty(cContain)
 62.                if "."$aFilename(nFor,1)
 63.                     cPrgname = substr(aFilename(nFor,1),1,;
 64.                          at(".",aFilename(nFor,1)) - 1)
 65.                else
 66.                     cPrgname = aFilename(nFor,1)
 67.                endif
 68.                
 69.                if not trim(cContain)$cPrgname
 70.                     lFile = .F.
 71.                endif
 72.           endif
 73.           
 74.           *output the directory name 
 75.           if lDir and lFile
 76.                ? cCurdir
 77.                lDir = .F.
 78.           endif
 79.           
 80.           *output the file name and append into database
 81.           if lFile
 82.                ? space(5) + aFilename(nFor,1)
 83.                append blank
 84.                replace Fname with aFilename(nFor,1), ;
 85.                          Fsize with aFilename(nFor,2), ;
 86.                          Fdate with aFilename(nFor,3), ;
 87.                          Ftime with aFilename(nFor,4), ;
 88.                          Fdir  with cCurdir
 89.           endif
 90.                               
 91.      endif
 92. endfor
 93. 
 94. release aFilename   &&free up memory


All subdirectories subordinate to the directory we just searched
will now be searched.  We accomplish this by calling FileFind
again.  If we were to watch this program push and pop itself off
the stack, we might be reminded of the game "Towers Of Hanoi".

 95. for nFor = 1 to nRowdir
 96. if substr(aDirname(nFor,5),5,1) = "D" and ;
 97.           substr(aDirname(nFor,1),1,1) # "."
 98.           cNextdir = cCurdir + aDirname(nFor,1) + "\"
 99.           do filefind with cNextdir
100.      endif     
101. 
102. endfor
103. return


The files that met our search pattern are now displayed to us in
a browse window.   The procedure ViewFile is called whenever the
operator presses the ENTER key. 

104. procedure viewfile
105. cFilename = trim(Fdir) + Fname
106. modify command &cFilename noedit
107. return


The FoxPro errors are waiting if a recursive process is called
without considering the limitations.  If planned properly, you
can create some nifty utilities.



     